// PROJECT : ALIB
// FILE    : ALHND.H - header file for handles
// AUTHOR  : Michael Vorburger, alpha ware, CH-SG
// DATE oC : January 1994
// DATE oLM: March 1995
// LEVEL   : GLE

#ifndef __ALHND_H
#define __ALHND_H

#include <error.h>
#include <debug.h>

#ifndef __ALOBJECT_H
#  include <alobject.h>
#endif

/* ------------------------------------------------------------------------ */

// FORWARD DECL
template <class T> class Shell;

/* ------------------------------------------------------------------------ *\
    CLASS : Fruit
    TYPE  : ABC, belongs to my Shell/Fruit-Pattern
    DESC  : This class has to be mixed-in in any other class that will be
            used as a FRUIT in connection with the Shell<> template.
    REMs  :

// Fruit nur als "Frucht" fr alShellTempl verwenden!
// ALL Fruit-Objects MUST be dynamic, NOT static!
// NEVER delete an Fruit!

// MAKRO  : Fr jede Fruit-Klasse K muss genau 1x das FRUITMACRO im
// public-Teil auftauchen. Siehe Bemerkung unten. Setzt einen
// Copy-Konstruktor (ev. autom. gen.) voraus.
// Gilt _NICHT_ fr ABCs.
*/
#define FRUITMACRO(F) Fruit& Copy() const { return ref_cast(F,*new(F(*this))); };
/*
#define FRUITMACRO(F) \
   public:    Fruit& Copy() const { return ref_cast(F,*new(F(*this))); }; \
   private: void operator delete(void*);                                \
   public:
*/
// NB: The new C++ std draft will allow the following (instead the ref_cast):
//      F& Copy() const { return *new F(*this)); };


class Fruit : public AlBObject
{public:
   Fruit()               : refs(0) {};
   Fruit(const Fruit& C) : refs(0) {}; // _NICHT_ refs(C.refs) !!!


   void MakeStatic() const     { const_cast(Fruit* const, this)->refs=-2; };
/* Erklrung: Fruits MSSEN eigt. immer dynamische Heap-Objekte sein. In
   Ausnahmefllen ist es aber erlaubt, sie statisch zu machen. Dann muss aber
   diese Methode fr das Objekt aufgerufen werden. NY() und YAF() werden dann
   ignoriert.
   Diese Sache ist noch nicht vollstndig & sicher berdacht... ev. mssen
   solche Fruits immer const sein??                                         */

 protected:
   virtual ~Fruit()
	 { if ((refs--)>0)
	     Error(ErrWrongDelete);
       refs=-1;                };

   virtual Fruit& Copy() const = 0;  // Siehe Makro oben.

 public: // Msste <private:> sein. Hilft aber alles nix; Shell kann mit
         // diesem Compiler nicht zum friend von Fruit gemacht werden. su.
   SWORD refs;		// How many references to this object?
                    // special cases: -1 = dead object, -2 = static object
   void NY() const  // this func is called (by Shell) if sbd points to this object.
	 { if(refs>-1)  // special cases: -1 = dead object, -2 = static object
         const_cast(Fruit* const, this)->refs++;
       else if(refs==-1)
         Error(ErrAccDelObj); };

   void YAF() const   // means: YouAreFree. YAF is called if the object isn't
	 { if (refs>-2) { // needed anymore. If nobody wants me, I'll deleted.
         if (refs<1)     // by Shell.
           Error(ErrTooManyYAF);
		 else
		   const_cast(Fruit* const, this)->refs--;  }; };

   // virtual??? MUST BE! - NO, see trick in Shell::operator==();
   bool operator==(const Fruit& C) const; // This func has NO DEFINITION. That's ok.
   // If a user wants to compare Hnd (say: also Fruits) he must redefine an
   // operator==(const DerrivedFruitType&) later. If a user compares
   // Hnds<DerrF> without redefining DerrF::operator==(DerrF), the
   // _linker_ generates an error.

   friend class ShellEngine;

   friend class Shell<Fruit>; // diese Zeile bewirkt _nix_ so. Siehe oben.
   //darum weglassen; bool Shell::operator==(const Shell<T>& c) const geht
   //sonst nicht.

// RELICT
//   virtual BOOL IsA(UWORD id) const
//     { return( id == IdFruit ); };
//   virtual UWORD TypeOf() const
//     { return IdFruit; };
};


// -*-*-*-**-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
// Shell<T> : Ein "intelligenter" Pointer. T muss eine Fruit sein.
// DON'T derrive classes from Shell more than once and don't use use pointers
// to any "general Shell" and don't override member and ...
// Always derrive publicly from Shell.
// MAKROS:
// SHELL_INTERFACE(self.Typ, FruitTyp) genau 1x pro Shell.
#define DERR_SHELL_INTERFACE_CSTR(S,P,F) \
        S()                        : P()    { }; \
        S(const F& arg)            : P(arg) { }; \
        S(const Shell<F>& arg)     : P(arg) { }; \
    bool operator==(const S& c) const            \
       { if(ShellEngine::operator==(c)) return true;       \
         else return ref_cast(const F,FrtConst())==ref_cast(const F,c.FrtConst()); }; \
    bool operator!=(const S& c) const            \
       { if(ShellEngine::operator==(c)) return false;      \
         else return !( ref_cast(const F,FrtConst())==ref_cast(const F,c.FrtConst()) ); };
//        S(const F* arg)            : P(arg) { }; \
//        S(const Shell<F>* arg)     : P(arg) { }; \

#define DERR_SHELL_INTERFACE_ASSG(S,P,F) \
        P& operator=(const F& arg) { EqualFrtPtr(&arg); return *this; }; \
        P& operator=(const S& arg) { EqualShell(arg); return *this;}; //\
//        Shell<F>& operator=(const ShellEngine& S)  { ShellEngine::EqualShell(S); return *this; };
//        Shell<F>& operator=(const F* arg) { return EqualFruitPtr(arg);  }; \
//        Shell<F>& operator=(const Shell<F>* arg) { return EqualShell(*arg);    }; //\

#define DERR_SHELL_INTERFACE(S,P,F) DERR_SHELL_INTERFACE_CSTR(S,P,F) \
                                    DERR_SHELL_INTERFACE_ASSG(S,P,F)

#define SHELL_INTERFACE(S,F) DERR_SHELL_INTERFACE(S,Shell<F>,F)
// Diese Makros braucht es, weil Konstruktoren und operator= nicht "richtig"
// vererbt werden.
    //#if (sizeof(S)!=sizeof(Shell<F>)) \
    //  #error Uops! Your handle has additional data members or virtual members \
    //#endif

// FRUIT_INTERFACE(self.Typ, FruitTyp, InitializerTyp) bel. viele Male. Die
// Shell kann dann mit InitializerTyp initialisiert werden. Achtung: Die
// Entsprechende Fruit braucht einen Konstruktor UND EINEN operator=() [!!]
// mit Arg InitTyp.
#define DERRSH_FRTINI_INTF_FAM(S,SP,F,FP,T) \
        S(const T  init)                   : SP( *new F(init)  )       {  }; \
        Shell<FP>& operator=(const T  arg) { return EqualFrtPtr(new F(arg)); }; //\
//        S(const T* init)                   : SP( new F(*init)  )      {  }; \
//        Shell<FP>& operator=(const T* arg) { return EqualFruitPtr(new F(*arg));};
#define DEC_DERRSH_FRTINI_INTF_FAM(S,SP,F,FP,T) \
        S(const T  init)                   ; \
        Shell<FP>& operator=(const T  arg) ;
#define DEF_DERRSH_FRTINI_INTF_FAM(S,SP,F,FP,T) \
        S::S(const T  init)                : SP( *new F(init)  )       {  }; \
        Shell<FP>& S::operator=(const T  arg) { return EqualFrtPtr(new F(arg)); }; //\

#define FRUITINIT_INTERFACE_FAM(S,F,FP,T) \
        DERRSH_FRTINI_INTF_FAM(S,Shell<FP>,F,FP,T)
#define DEC_FRUITINIT_INTERFACE_FAM(S,F,FP,T) \
        DEC_DERRSH_FRTINI_INTF_FAM(S,Shell<FP>,F,FP,T)
#define DEF_FRUITINIT_INTERFACE_FAM(S,F,FP,T) \
        DEF_DERRSH_FRTINI_INTF_FAM(S,Shell<FP>,F,FP,T)

#define DERRSHELL_FRUITINIT_INTERFACE(S,SP,F,T) \
        DERRSH_FRTINI_INTF_FAM(S,SP,F,F,T)
#define DEC_DERRSHELL_FRUITINIT_INTERFACE(S,SP,F,T) \
        DEC_DERRSH_FRTINI_INTF_FAM(S,SP,F,F,T)
#define DEF_DERRSHELL_FRUITINIT_INTERFACE(S,SP,F,T) \
        DEF_DERRSH_FRTINI_INTF_FAM(S,SP,F,F,T)

#define FRUITINIT_INTERFACE(S,F,T) \
        FRUITINIT_INTERFACE_FAM(S,F,F,T)
#define FRUITINIT_INTERFACE_DEC(S,F,T) \
        DEC_DERRSH_FRTINI_INTF_FAM(S,F,F,T)
#define FRUITINIT_INTERFACE_DEF(S,F,T) \
        DEF_FRUITINIT_INTERFACE_FAM(S,F,F,T)
/*
// If F cannot be defined at the moment of declaration, use this:
#define FRUITINIT_INTERFACE_FAM_DEC(S,F,P,T) \
        S(const T  init); \
        Shell<P>& operator=(const T  arg); //\
//        S(const T* init); \
//        Shell<P>& operator=(const T* arg);
#define FRUITINIT_INTERFACE_FAM_DEF(S,F,P,T) \
        S::S(const T  init)                         : Shell<P>( *new F(init)  )    {  }; \
        Shell<P>& S::operator=(const T  arg) { return EqualFrtPtr(new F(arg)); }; //\
//        S::S(const T* init)                  { EqualFruitPtr(new F(*init));}; \
//        Shell<P>& S::operator=(const T* arg) { return EqualFruitPtr(new F(*arg));};
*/
class ShellEngine
{ public:
    void Reset() { Liberare(); };

    operator bool() const
      { if (r!=NULL)
          return FrtConst();
        else
          return false;  };

	ShellEngine()                        : r(NULL)                {                          };
	ShellEngine(const ShellEngine& arg)  : r(arg.r)               { arg.r->NY();             };
    ShellEngine(const ShellEngine* arg)  : r(arg->r)              { arg->r->NY();            };
    ShellEngine(const Fruit& arg)        : r(const_cast(Fruit*,&arg)) { arg.NY();                };
    ShellEngine(const Fruit* arg)        : r(const_cast(Fruit*,arg))  { if(arg!=NULL) arg->NY(); };
    DERR_SHELL_INTERFACE_ASSG(ShellEngine,ShellEngine,Fruit)

	~ShellEngine() { Liberare(); }; // no virtual please. saves memory.

  protected:
    const Fruit& FrtConst() const
    { if (r==NULL)
		{ Error(ErrNotInitVar); /* return ErrT; s.u. */ };
	  // else
		return *r;	};

    Fruit& FrtUpd()
      { Fruit& newFrt=FrtCopy();
        EqualFrtPtr(&newFrt);
        return newFrt;              };

    Fruit& FrtMustCpy() const // MyFrtMustCopy
    { if (r==NULL)
		{ Error(ErrNotInitVar); /* return ErrT; s.u. */ };
	  //else
          return ref_cast(Fruit,r->Copy());
          // FALSCH wre: *new Fruit(*r); weil so Probleme mit verberbten Fruit's
          // ref_cast wg. prob bei ndern von Copy() return type (so.)
	};

    ShellEngine& EqualFrtPtr(const Fruit* arg)
    { if (arg!=r) {
        if (arg!=NULL)
          const_cast(Fruit*,arg)->NY();
        Liberare();
        r=const_cast(Fruit*,arg);
      };
      return *this;                                                  };

    ShellEngine& EqualShell(const ShellEngine& arg)
	{ if(arg.r!=r) {
        arg.r->NY();
	    Liberare();
	    r=arg.r;
      };
	  return *this;	                                                 };

    bool operator==(const ShellEngine& c) const
      { if(r==c.r) return true;
        else       return false; };

private:
    Fruit& FrtCopy() const
    { if (r==NULL)
		{ Error(ErrNotInitVar); /* return ErrT; s.u. */ };
      // else {
		if(r->refs>1 || r->refs==-2)
          return ref_cast(Fruit,r->Copy()); // so.
        else
          return *r; /* }; */                                        };

 protected:
   inline void Liberare()
	  { if (r!=NULL)
		{ r->YAF();
		  if ( (*r).refs==0 )
		    delete r;
		  r=NULL;             };
	  };

 private:
   Fruit* r;

 // a) friend template <class T> class Shell;
 // b) friend class Shell;
 // c) friend Shell<T>;
 //friend class Shell<Fruit>;
};
/* REMARKS:
    1. Einig Methoden von Shell sollten ein Err-Objekt zurckliefern knnen.
    Idee: template <class T, T ErrT> und Shell<TermFruit, ErrTerm>
    Das geht leider nicht; obwohl zB template <class T, int sz> erlaubt ist.
*/

template <class T>  // T must _be_ a Fruit (publicly derrived from it)
class Shell : public ShellEngine
{   typedef ShellEngine E;
  public:
    void Reset()            { Liberare(); };

    operator bool() const   { return E::operator bool(); };
/*
    bool operator==(const Shell<T>& c) const
       { if(E::operator==(c)) return true;
         else return ref_cast(const T,FrtConst())==ref_cast(const T,c.FrtConst()); };
    bool operator!=(const Shell<T>& c) const
       { if(E::operator==(c)) return false;
         else return !( ref_cast(const T,FrtConst())==ref_cast(const T,c.FrtConst()) ); };
*/
// ATTENTION: The following methodes have to be defined here (in this class)
// because one might use type Shell<Fruit> directly -without derriving from
// it- AND ALSO in the derrived class (see macro SHELL_INTERFACE)
   DERR_SHELL_INTERFACE_CSTR(Shell,ShellEngine,T)
   Shell(const ShellEngine& se): ShellEngine(se)  {};
   DERR_SHELL_INTERFACE_ASSG(Shell,Shell,T)

  protected:
    T& Frt()                { return ref_cast(T,E::FrtUpd());         };
    const T& Frt()  const   { return ref_cast(const T,E::FrtConst()); };

    Shell<T>& EqualFrtPtr(const T* arg)    { E::EqualFrtPtr(arg); return *this;  };
    Shell<T>& EqualShell(const Shell& arg) { E::EqualShell(arg);  return *this;  };
    T&        FrtMustCpy() const           { return ref_cast(T,E::FrtMustCpy()); };

// the following is for backward compatibility, use the substitutes
    const T& MyFrtConst() const           { return Frt(); };
    Shell<T>& EqualFruitPtr(const T* arg) { return EqualFrtPtr(arg); };

 protected:
/*	virtual BOOL IsA(UWORD id) const
	{ return( id == IdShell
//                 || T::IsA(id)   Nicht machen! Eine Shell ist KEINE Fruit.
                             ); };

	virtual UWORD TypeOf() const
	  { return IdShell; };        */
};
// typedef Shell Hnd;
#define Hnd Shell
// #define AlShellTempl Shell

// SPECIAL VERSIONS OF FRUITINIT:
// Fr abgeleitete Klassen F von P (Parent), die mit einem T initialisiert
// werden knnen:

template<class T> class TrivialFruit : public Fruit
{  T item;
 public:
   TrivialFruit(const T& init)            : item(init)  {             };
   TrivialFruit& operator=(const T& init) { item=init;  return *this; };
//   TrivialFruit(const T* init)            : item(*init) {             };
//   TrivialFruit& operator=(const T* init) { item=*init; return *this; };

   //operator T() const                     { return item;             };
   operator const T&() const              { return item;             };
   //operator       T()                     { return item;             };
   operator const T*() const              { return &item;            };
   //operator       T*()                    { return &item;            };
   FRUITMACRO(TrivialFruit<T>)
};

// -*-*-*-**-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
// THndP<T> : TrivialHandle for primitive types (int, char etc)
// For direct using. You don't have to derrive a class from THndP before you
// can use it. It works like a pointer type. For exampe: THnd<int> I=new int;
//
// Always derrive publicly from THnd.

template<class T> class THndP : public Hnd< TrivialFruit<T> >
{  typedef TrivialFruit<T> TFruit;
 public:
    const T& operator *()  const  { return Frt();  };
	T        operator* ()         { return Frt();  };

    const T& operator() () const  { return Frt();  };
	T       operator() ()         { return Frt();  };

   SHELL_INTERFACE(THndP,TFruit)
   FRUITINIT_INTERFACE(THndP,TFruit,T)
   //FRUITINIT_INTERFACE(THndP,TFruit,T*)

	//operator const T&() const     { return FrtConst();  };
	//operator T&()                 { return FrtUpd();    }; // ???
// Der cast-Operator geht nicht immer; Compiler muss WISSEN, welcher Typ
// bentigt wird. Bsp. Geg. F f; S<F> s; class F has method m();
// void func(const F& a); f.m() [ok]; s.m() [no!]; func(f) [ok]; func(s) [ok];

// >> Forget about this auto-cast forever, ok? I decided that my Handles don't
// support it! It doesn't work because compiler cannot distinguish these two
// operators well enough.
};

// THndC : TrivialHandle for class types
// derrived from THndP, additionally supports operator->
template<class T> class THndC : public THndP<T>
{  typedef TrivialFruit<T> TFruit;
 public:
    const T* operator-> () const  { return Frt(); };
    T*       operator-> ()        { return Frt(); };

    DERR_SHELL_INTERFACE(THndC,THndP<T>,TFruit)
    DERRSHELL_FRUITINIT_INTERFACE(THndC,THndP<T>,TFruit,T)
    //DERRSHELL_FRUITINIT_INTERFACE(THndC,THndP<T>,TFruit,T*)
};

#endif	// __ALHND_H

