// PROJECT : ALIB
// FILE    : ALSTRING.H - header file for ALIB strings
// AUTHOR  : Michael Vorburger, alpha ware
// DATE oC : August 1994
// DATE oLM: August 1994
// LEVEL   : GLE


#ifndef __ALSTRING_H
#define __ALSTRING_H

#include <alobject.h>

// #define UseStringLib
#if defined(UseStringLib)   // Verwende Str++ 3.0 (download von Fido) ?
#include <str.h>

#else                       // Nein, dann ALIB
#include <string.h>
#include <stdlib.h> // for itoa()

// #define strDEBUG         // Debugging fr Strings

/* So n Saich; das untenstehende wre doch eine saubere Lsg fr
   Debug-Ausgaben-Ifdefs. Statt immer
     #ifdef strDEBUG
         cerr << "blabla" << endl;
     #endif
   knnte man dann schreiben
      sDbg cerr << "blabla" << endl;

#ifdef strDEBUG
#define sDbg
#else
#define sDbg //     <- das geht nicht!
#endif */

// machen wir's halt so:
#ifdef strDEBUG
#define DBGPRT(m) cerr << m << endl;
#else
#define DBGPRT(m)
#endif


class StringFruit : public Fruit
{ private:
    char* s;

   public:
     FRUITMACRO(StringFruit)

     StringFruit(char* i)
      { s=new char[strlen(i)+1]; strcpy(s,i);
        DBGPRT("StringFruit konstruiert aus char* = \"" << s << "\"")
      };
    StringFruit(const char* i)
      { s=new char[strlen(i)+1]; strcpy(s,i);
        DBGPRT("StringFruit konstruiert aus CONST char* = \"" << s << "\"")
      };

//     StringFruit(const char i)
//      { s=new char[2]; s[0]=i; s[1]=0;
//        DBGPRT("StringFruit konstruiert aus const char = '" << s << "'")
//      };

/* StringFruit(const char* i)
      { s=(char*)i; };
    // Das ist natrlich etwas frech, einfach zu casten. Ich denke aber
    // hier ist es erlaubt und sinnvoll. Denn ... ???

   Nein, ist es nicht! Folgender Code:
                                        String      s="Mike";
     Beim Umweg ber einen const        const char* cp=s;     // keine Kopie!
     char* geht die Information
     verloren, ob der String schon      String      e(cp);
     benutzt wird. Das ist beim         // StringFruit(const char* i) benutzt!
     const char* noch egal, nicht
     mehr aber bei einem neuen String() e[0]='D';
                                        cout << s; // gibt "Dike" aus!       */
    StringFruit(const StringFruit& i)
      { s=new char[strlen(i.s)+1]; strcpy(s,i.s);
        DBGPRT("StringFruit konstruiert ber Copy-Konstruktor von \"" << s << "\"") };

    StringFruit(const SWORD i)
      { char num[18];
        itoa(i, num, 10);
        s=new char[strlen(num)+1]; strcpy(s,num);
        DBGPRT("StringFruit konstruiert aus SWORD(" << i << ")") };

    ~StringFruit() {
      DBGPRT("StringFruit \"" << s << "\"" << " destruiert!")
      delete[] s;
      s=0;         };

    StringFruit& operator+ (const char* c)
       { char* t=new char[strlen(s)+strlen(c)+1];
         DBGPRT("StringFruit \"" << s << "\" operator+(\"" << c << "\")")
         strcpy(t, s);
         strcpy(&t[strlen(s)] ,c);
         delete[] s;
         s=t;
         return *this; };

    StringFruit& operator+ (const char C)
       { char* t=new char[strlen(s)+1+1];
         DBGPRT("StringFruit \"" << s << "\" operator+(\"" << C << "\")")
         strcpy(t, s);
         t[strlen(s)]=C; t[strlen(s)+1]=0;
         delete[] s;
         s=t;
         return *this; };

    virtual BOOL IsA(UWORD id) const
		{ return( id==IdStringFruit || Fruit::IsA(id) ); };
	virtual UWORD TypeOf() const
		{ return IdStringFruit; };

  friend class String;
/* Wenn der Compiler diese (richtige!) Zeile schnallen wrde, knnte oben
   der Destruktor auch noch private gemacht werden.
//  friend class Shell<StringFruit>;                                      */
};

class String : private Shell<StringFruit>
{ // String wird von Shell vllig entkoppelt. Meyers sagt: Wir erben nur die
  // Implementation, jedoch nichts von der Schnittstelle. -- Warum nicht ber
  // Layering gelst, siehe r41.
  public:

    SHELL_INTERFACE(String,StringFruit)
/*
#define S String
#define F StringFruit
        S()                               { }; \
	    S(const S& arg)                   { EqualShell(arg);           }; \
        S(const S* arg)                   { EqualShell(*arg);          }; \
        S(F& arg)                         { EqualFruitPtr(&arg);       }; \
        S(F* arg)                         { EqualFruitPtr(arg);       }; \
        Shell<F>& operator=(F& arg)       { return EqualFruitPtr(&arg);}; \
        Shell<F>& operator=(F* arg)       { return EqualFruitPtr(arg); }; \
        Shell<F>& operator=(const S& arg) { return EqualShell(arg);    }; \
        Shell<F>& operator=(const S* arg) { return EqualShell(*arg);   };
*/
    FRUITINIT_INTERFACE(String,StringFruit,char*)
    FRUITINIT_INTERFACE(String,StringFruit,const char*)
    FRUITINIT_INTERFACE(String,StringFruit,const SWORD)

    operator const char*() const
      { return MyFrtConst().s; };
    operator char*() const
      { return MyFrtMustCopy().s; };
/* the methode above was like this:
    operator char*()
      { char* t=new char[strlen(MyFrtConst().s)+1];
        strcpy(t,MyFrtConst().s);
        return t; };
*/
    const char& operator[] (int i) const
      { CheckPosRange(i);
        return MyFrtConst().s[i]; };
    char& operator[] (int i)
      { CheckPosRange(i);
        EqualFruitPtr(&MyFrtCopy());
        return MyFrtConst().s[i]; };

    StringFruit& operator+ (const char* a) const
      { return (MyFrtMustCopy()+a); };
    void operator+=(const char* a)
      { EqualFruitPtr( &(MyFrtCopy()+a)); };
    StringFruit& operator+ (const String& a) const
      { return (MyFrtMustCopy() + a.MyFrtConst().s); };
    void operator+=(const String a)
      { EqualFruitPtr( &(MyFrtMustCopy() + a.MyFrtConst().s)); };
    StringFruit& operator+ (const char c) const
      { return (MyFrtMustCopy()+c); };
    void operator+= (const char c) 
      { EqualFruitPtr( &(MyFrtMustCopy()+c)); };

    bool operator== (const char* c) const
      { return !strcmp(MyFrtConst().s,c); };
    bool operator== (const String& c) const
      { return !strcmp(MyFrtConst().s,c.MyFrtConst().s); };

    bool operator> (const char* c) const
      { return(strcmp(MyFrtConst().s,c) > 0); };
    bool operator> (const String& c) const
      { return(strcmp(MyFrtConst().s,c.MyFrtConst().s) > 0); };
    bool operator< (const char* c) const
      { return(strcmp(MyFrtConst().s,c) < 0); };
    bool operator< (const String& c) const
      { return(strcmp(MyFrtConst().s,c.MyFrtConst().s) < 0); };

    void Reset() { Shell<StringFruit>::Reset(); };

    friend istream& operator>> (istream& i, String& s);
/*
    virtual BOOL IsA(UWORD id) const
		{ return( id==IdString || Shell<StringFruit>::IsA(id) /*|| isASCIIable::IsA(id)*/ ); };
	virtual UWORD TypeOf() const
		{ return IdString; };
*/
 private:
    void CheckPosRange(int i) const
      { if( i<0 || i>strlen(MyFrtConst().s) ) Error(ErrString); };
};

String operator+ (const char* a, const String& b);

#endif // defined(UseStringLib)

class isASCIIable //: public AlObject
{ public:
	virtual String AsASCII() const = 0;

    operator String() { return AsASCII(); };
	virtual BOOL IsA(UWORD id) const
		{ return(id==IdisASCIIable); };
	virtual UWORD TypeOf() const
	 { return IdisASCIIable; };
};
extern ostream& operator<< ( ostream&, const isASCIIable&);

#endif // __ALSTRING_H