#include <alobject.h>
#include <alstring.h>
#include <portab.h>
//#include <error.h>

/* ------------------------------------------------------------------------ */
// ListBase
// Eine ListBase enthlt ein dyn. Array von Elementen. Sie ist nur fr
// elementare Typen mit trivialer oder automatischer Speicherverwaltung
// geeignet.
// Die Elemente werden unter Umstnden umkopiert; also mglichst kleine
// Typen whlen. ACHTUNG: Nie Listen von Zeigern verwenden; dazu sind
// Shells zu verwenden!

template<class T> class ListBase : public AlObject //, public debugable
{ protected:
    UWORD size;
	UWORD incr;
	UWORD next;

                  // WAS: typedef T* TMemArr;
    T* pMemArr;   // WAS: T**

  public:
	ListBase(UWORD startSize = 10, UWORD increase = 5);
    ListBase(const ListBase& copyL);
    ListBase& operator= (const ListBase& copyL);
	virtual ~ListBase();

    void      Append(const T& e);
    void      Insert(UWORD p, const T& e);
    void      Insert(LIter<T> p, const T& e){ p.ValidFor(*this); Insert(p.i, e); };
    void      Remove(UWORD p);
    void      Remove(LIter<T> p)            { p.ValidFor(*this); Remove(p.i);    };
    ListBase& operator+=(T& a)	            { Append(a); return *this;           };

    T& operator[] (UWORD p) const           // [] gibt das x. Element zurck
       { if ( ( (p-1)<next) && p>0 )        // Zhlung ab 1 !!
           return pMemArr[p-1];
         else
           Error(ErrOORinLA); };
    T& operator[] (LIter<T> p) const        { p.ValidFor(*this);
                                              return operator[] (p.i);          };
    UWORD     Size() const                  { return next;                      };

    BOOL IsA(UWORD id) const { return id==IdList; };
	UWORD TypeOf() const     { return IdList; };
};

template<class T> ListBase<T>::
ListBase(UWORD startSize = 10, UWORD increase = 5)
  {
    pMemArr=new T[startSize];
    if(!pMemArr)
      Error(ErrNoMoMemory);
    size=startSize; incr=increase; next=0;
   };

template<class T> ListBase<T>::
~ListBase()
  { delete[] pMemArr; };

template<class T> ListBase<T>::
ListBase(const ListBase<T>& copyL)
  { size=copyL.size; incr=copyL.incr;
    pMemArr=new T[size];
    for(next=0; next<copyL.next; next++)
      pMemArr[next]=copyL.pMemArr[next];
  };

template<class T>
ListBase<T>& ListBase<T>::operator= (const ListBase<T>& copyL)
  { delete[] pMemArr;
    size=copyL.size; incr=copyL.incr;
    pMemArr=new T[size];
    for(next=0; next<copyL.next; next++)
      pMemArr[next]=copyL.pMemArr[next];
    return *this;
  };

template<class T> void ListBase<T>::
Append(const T& e)
{ if(next >= size) {
    T* old=pMemArr;
    size+=incr;
    pMemArr=new T[size];
    if(!pMemArr)
      Error(ErrNoMoMemory);
    for(UWORD i=0; i<next; i++)
      pMemArr[i]=old[i];
    delete[] old;
   };

   pMemArr[next++]=e;
};

template<class T> void ListBase<T>::
Insert(UWORD p, const T& e)
{  if (!( ( (p-1)<next) && p>0 ))
     Error(ErrOORinLA);

   if(next >= size) {
     T* old=pMemArr;
     size+=incr;
     pMemArr=new T[size];
     if(!pMemArr)
       Error(ErrNoMoMemory);

     UWORD i;
     for(i=0; i<(p-1); i++)
       pMemArr[i]=old[i];
     for(i=next; i>(p-1); i--)
       pMemArr[i]=old[i-1];
     delete[] old;
   }
   else {
     for(UWORD i=next; i>(p-1); i--)
       pMemArr[i]=pMemArr[i-1];
     next++;
   };

   pMemArr[p-1]=e;
};

template<class T> void ListBase<T>::
Remove(UWORD p)
{  if (!( ( (p-1)<next) && p>0 ))
     Error(ErrOORinLA);

   for(UWORD i=(p-1); i<(next-1); i++)
     pMemArr[i]=pMemArr[i+1];

   next--;
};

/* ------------------------------------------------------------------------
// Iterator
   Mit einem Iterator kann auf eine Liste/Menge etc. zugegriffen werden. Er
   ersetzt bzw. erweitert die Mglichkeiten des alten Integer-Indizes.
   Ein Iterator ist Listenspezifisch (nicht austauschbar) und kann seine
   Gltigkeit verlieren; siehe SetBase.

   NOCH NICHT VOLLSTNDIG UND KONSEQUENT IMPLEMENTIERT.
*/

/*
class Iterator {
  Iterator(
*/

template<class T>
class LIter
{ const ListBase<T>* L;
  UWORD     i; // Index im Format 1..N nicht 0..N
public:
  LIter(const ListBase<T>& l)
    : L(&l), i(1)
    { if(L->Size()==0)
       i=0;        };
  bool Valid()                          { return (( i<=L->Size()) && i>0);};
  bool ValidFor(const ListBase<T>& c)   { if(L!=&c)
                                            Error(ErrWrongIter);
                                          return (L==&c);                };
  void operator++ ()                    { i++; if (!Valid()) i=0;          };
  operator bool()                       { return Valid();                };

  friend ListBase<T>;
};

class BIter {
  UWORD age;

public:
  bool Valid(); // Ist der Iterator noch gltig?
  operator bool() { return Valid(); };
};

/* ------------------------------------------------------------------------
// SetBase
   Eine SetBase (Menge) ist eigt. eine sortierte Liste, die jedes Element
   nur einmal enthlt. Sie bietet folgende Methoden:
        - Add(T) oder +=(T) fgt der Menge ein Element hinzu. Wenn schon
           vorhanden, passiert NIX; das ist kein Fehler.
        - Find(T) "sucht" ein Element mittels der Elementmethode
           T::Match(const T&) den alle Elemente zur Verfgung stellen mssen.
           Find() liefert eine Variable vom Typ BIter zurck; der aber nur bis
           zum nchsten Add() oder Remove() Gltigkeit hat.
        - Get(BIter) liefert das zu BIter gehrende Elemente zurck. Su.
        - Remove(BIter) entfernt das zu BIter gehrende Element aus der Menge.
   Die Elemente mssen die Operator > und = und != und < untersttzen.

   Mengen (SetBase) knnen mit | vereinigt und mit & geschnitten werden.
*/

// NOCH NICHT VOLLSTNDIG IMPLEMENTIERT!


template<class T> class Set : public AlObject //, public debugable
{ protected:
    ListBase<T>& L;
    UWORD age;      // Zugriffskontrolle fr BIter

  public:
    void NOT_USABLE() = 0;

    Set(UWORD startSize = 10, UWORD increase = 5)
     : L(*new ListBase(startSize, increase), age(0)
     { };
    Set(const Set& c);
    operator= (const Set& c);
    virtual ~Set() { };

    void Add(const T& e)
     { LIter<T> I(L);
       while(I)         // I heisst I.Valid()
         if(e>L[I])
           I++;
       if(e!=L[I]) {
         if(I)
           L.Insert(I,e);
         else
           L.Append(s4);  };
     };

    T&    Get(BIter i);
    void  Remove(BIter i);

    BIter Find(const T& e)  // Intervallhalbierungs-Algorithmus
     { LIter<T> I(L);
       I.Set(L.Size()/2);

     };
};

// ============================================================================

/* ------------------------------------------------------------------------ */
// alte Herangehensweise: List enthlt Zeiger auf dyn. Array von Zeigern
//                        auf die Elemente. Problem wg. Freigabe etc.
/*
template<class T> class ListBase : public AlObject //, public debugable
{ protected:
    UWORD size;
	UWORD incr;
	UWORD next;

    // typedef T* TMemArr;
    T** pMemArr;

  public:
	ListBase(UWORD startSize = 10, UWORD increase = 5);
    ListBase(ListBase& copyL);
	virtual ~ListBase();

    void Append(T& e);
    void Insert() {};
    void Remove() {};


// ListBase& operator+= (T& a)	{ Add(a); return *this; };
    T& operator[] (UWORD i)
       { if ((i-1)<next)
           return *pMemArr[i-1];
         else
           Error(ErrOORinLA); };

    BOOL IsA(UWORD id) const { return id==IdList; };
	UWORD TypeOf() const     { return IdList; };
};

template<class T> ListBase<T>::
ListBase(UWORD startSize = 10, UWORD increase = 5)
  {
    pMemArr=new TMemArr[startSize];
    if(!pMemArr)
      Error(ErrNoMoMemory);
    size=startSize; incr=increase; next=0;
   };

template<class T> ListBase<T>::
~ListBase()
 { delete pMemArr; };

template<class T> void ListBase<T>::
Add(T& e)
{ if(next >= size) {
    T** old=pMemArr;
    size+=incr;
    pMemArr=new TMemArr[size];
    if(!pMemArr)
      Error(ErrNoMoMemory);
    for(UWORD i=0; i<next; i++)
      pMemArr[i]=old[i];
    delete old;
   };

   pMemArr[next++]=&e;
};


template<class T> class ListDyn : public ListBase<T>
{ public:
    ~ListDyn()
      { for(UWORD i=0; i<next; i++)
          delete pMemArr[i];        };
    ListDyn(UWORD startSize = 10, UWORD increase = 5)
      : ListBase<T>(startSize, increase) { };
};

*/

/* ------------------------------------------------------------------------ */
/*  ListOS = List of Shells
    Statt Zeigern werden hier echte Shells in einem dynamischen Arrays
    gehalten. Shells belegen nur soviel Platz wie zwei Zeiger.


template<class T> class ListOF : public AlObject //, public debugable
{ protected:
    UWORD size;
	UWORD incr;
	UWORD next;

    // typedef T* TMemArr;
    T** pMemArr;

  public:
	ListBase(UWORD startSize = 10, UWORD increase = 5);
    ListBase(ListBase& copyL);
	virtual ~ListBase();

	// void Add(T& e);
    Insert
    Remove
    Append


//     ListBase& operator+= (T& a)	{ Add(a); return *this; };
    T& operator[] (UWORD i)
       { if ((i-1)<next)
           return *pMemArr[i-1];
         else
           Error(ErrOORinLA); };

    BOOL IsA(UWORD id) const { return id==IdList; };
	UWORD TypeOf() const     { return IdList; };
};
*/
/* ------------------------------------------------------------------------ */
