// typedef enum {false, true} bool; //
#include <portab.h>
#include <error.h>
#include <alstring.h>

/* ------------------------------------------------------------------------ */
// FORWARD-Deklarationen
template<class I, class O, class P> class Iter;
template<class T> class Array;
template<class T> class AIter;
template<class I, class O, class P> class OIter;

/* ------------------------------------------------------------------------ */
// Container<I,O,P> & Iter<I,O,P> - abstract base class for EVERYTHING

template<class I, class O, class P> class Container
{ public:
   virtual void      Add(const I&)                    = 0;
   virtual Iter<I,O,P>& Find(const I&) const          = 0;
   virtual O&        Get(const Iter<I,O,P>&)          = 0;
   virtual const O&  Get(const Iter<I,O,P>&) const    = 0;
   virtual void      Remove(const Iter<I,O,P>&)       = 0;

   Container& operator+=(const I& e)               { Add(e); return *this; };
   Container& operator<<(const I& e)               { Add(e); return *this; };
   O&       operator[](const Iter<I,O,P>& i)       { return Get(i);        };
   const O& operator[](const Iter<I,O,P>& i) const { return Get(i);        };

   virtual Iter<I,O,P>& NewIter() const = 0;

  friend class Iter<I,O,P>;
};


/* All classes derrived from Iter...
    1) must have a copy-construtor for type Iter.
    //2) should not allow to use operator= ; make it private.
*/
template<class I, class O, class P> class Iter
{ public:
    Iter(const Container<I,O,P>& c) : C(c) { };
    Iter& operator=(const Iter& s) { ValidFor(s.C); i=s.i; return *this; };

    virtual bool Valid() const = 0;

    const O& operator() () { return C.Get(*this); };
    operator bool() { return Valid(); };
    bool ValidFor(const Container<I,O,P>& c) const
      { if(&C!=&c) Error(ErrWrongIter); /*else*/ return true; };

 protected:
    const Container<I,O,P>& C;
    P     i;
    //union T_u u;
    friend class Array<I>;
};

/* ------------------------------------------------------------------------ */
// Set<T>
// A Set is an unordered set (:-) of items; so you cannot iterate through it!

template<class T> class Set : public Container<T, T>
{ public:

};

/* ------------------------------------------------------------------------ */
// OrdContainer<I,O,P> & OIter<I,O,P> - abstract base class
// An OrdContainer is an ordered thing: You can access the "third" item and
// you can iterate (by using UWORDs or OrdIters) through it.
template<class I, class O, class P> class OrdContainer : public Container<I,O,P>
{ public:
   Iter<I,O,P>& NewIter() const =0; // return type: OIter<I,O,P>&
   OIter<I,O,P>& NewIter_OIter() const
     { return (OIter<I,O,P>&)NewIter(); }; // TC!

   virtual O&        Get(const UWORD)          = 0;
   virtual const O&  Get(const UWORD) const    = 0;

   const O& operator[](const Iter<I,O,P>& i) const  { return /*i.*/Get(i); };
   const O& operator[](UWORD i) const               { return Get(i);       };
      //{ OIter<I,O,P> I(*this,i); return I.Get(); };
};

//enum OIterSetArg { FIRST, MIDDLE, LAST };
template<class I, class O, class P> class OIter : public Iter<I,O,P>
{   //OIter& operator=(const OIter& s);
  public:
    OIter(const OrdContainer<I,O,P>& c) : Iter<I,O,P>(c) { };
    OIter(const Iter<I,O,P>& J) : Iter<I,O,P>(J) { };
    // OIter(const OrdContainer<I,O,P>& c, UWORD p);
    virtual OIter& operator++ () = 0; // { i++;  if (!Valid()) i=0; };
    //void operator++ (int);
    virtual OIter& operator-- () = 0; // { i--;  if (!Valid()) i=0; };
    //void operator-- (int);
    //void Set(OIterSetArg p);
/*    { switch(p) {
      case FIRST : //i=1;          break;
      case MIDDLE: //i=L.Size()/2; break;
      case LAST  : //i=L.Size();   break;
      }
    };*/
};

/* ------------------------------------------------------------------------ */
// Array<T> & AIter<T>
// An Array behaves like a built-in array: It's size is fixed.

template<class T> class Array : public OrdContainer<T, T,UWORD>
{  typedef Iter<T,T,UWORD> itnAIter; // InTerNal IAIter
   void CheckBnd(UWORD j) const { if(j>(l-f)) Error(ErrArrayBounds); };
 public:
   Array(SWORD a, SWORD e=0);
   virtual ~Array() { delete[] arr; };

   void       Add(const T& i);
   itnAIter&  Find(const T&) const;
   T&        Get(const itnAIter& p)        { CheckBnd(p.i); return arr[p.i]; };
   const T&  Get(const itnAIter& p) const  { CheckBnd(p.i); return arr[p.i]; };
   T&        Get(const UWORD i)            { CheckBnd(i); return arr[i-f]; };
   const T&  Get(const UWORD i) const      { CheckBnd(i); return arr[i-f]; };
   void      Remove(const itnAIter&);

   virtual Iter<T,T,UWORD>& NewIter() const { return *new AIter<T>(*this); };

  protected:
    SWORD f, l; // first & last
    SWORD n;    // next item position for a new item
    T*    arr;  // ptr to real array

  friend class AIter<T>;
    // Iter<T,T> NewIter() const { return(*this,0); };
};

template<class T> class AIter : public OIter<T,T,UWORD>
{  typedef const Array<T>& rArrT; // cannot cast wg. virtuellen Basis-Klassen!
   #define aC TC(C,rArrT)
   typedef OIter<T,T,UWORD> Inhd;
   //AIter& operator=(const AIter& s);
  public:
    AIter(const Array<T>& c)          : OIter<T,T,UWORD>(c) { i=0; };
    AIter(const Array<T>& c, SWORD p) : OIter<T,T,UWORD>(c) { i=p-aC.f; aC.CheckBnd(i); };
    AIter(const Iter<T,T,UWORD>& I)   : OIter<T,T,UWORD>(I) { };
    const T& operator() ()                                  { return aC.arr[i]; };
    bool Valid() const                                      { return (i<(aC.l-aC.f)); };

    OIter<T,T,UWORD>& operator++ () { aC.CheckBnd(i); i++; };
    OIter<T,T,UWORD>& operator-- () { aC.CheckBnd(i); i--; };

  protected:
    // UWORD i;   // currently "iterated" item
    //AIter(const Array<T>& c, UWORD p) : OIter<T,T,UWORD>(c) { i=p; };
  friend class Array<T>;
};

/* ------------------------------------------------------------------------ */
// List<T> & LIter<T>
// Lists are similar to Arrays, but they are "unlimited"
template<class T> class List : public OrdContainer<T, T>
{ public:

};

/* ------------------------------------------------------------------------ */
// SortList<T> & sLIter<T>
// SortLists are similar to Lists, but all elements are automatically sorted
// Take care: You have no idea & control at which position an item is stored;
// but SortLists still are OrdContainers because you can iterate through them.
template<class T> class SortList : public OrdContainer<T, T>
{ public:
};

/* ------------------------------------------------------------------------ */
// Dictionary<K,V> & DIter<V>
// A dictionary contains two lists: One for storing a _Key and another for a
// _Value. You have to use Add(K,V) to add a pair of a key and a value which
// you will be able to access later by using Find(K) with a key. Find() will
// return an Iter; use it to get back the Value with V Get(Iter).
template<class K, class V> class Dictionary : public OrdContainer<K, V>
{ public:
    void      Add(const K&, const &V);
    Iter<K,V> Find(const I&) const;
    V&        Get(const Iter<K, V>&);
    const V&  Get(const Iter<K, V>&) const;
    void      Remove(const Iter<L, V>&);

    void Add(const K&); // NOT defined! If used anyway, the linker detects it.
};

/* ------------------------------------------------------------------------ */
// ContUWI<ItemT,ContainerT> - ContainerUserWorldInterface
// Derrive any Container class from UWI_O and it can be outputet to cout!
/*
template<class T> class ContUWI : virtual public OrdContainer<T,T,SWORD>, public isASCIIable {
  String AsASCII() const
   { OIter<T,T,SWORD>& I=NewIter_OIter();
     String s("{");
     while(I)
       { s+=I(); I++; };
     s+="}";
     return s;             };
};
Das funktioniert aus div. Grnden (im Moment?) nicht wie gedacht...
*/

/* ------------------------------------------------------------------------ */
//                            IMPLEMENTATION
/* ------------------------------------------------------------------------ */

template<class T> Array<T>::
Array(SWORD a, SWORD e=0) {
  if(a<e && e!=0)
    { f=a; l=e; }
  else if(a>0 && e==0)
    { f=0; l=a; }
  else
    Error(ErrArrayBounds);

  arr=new T[l-f+1];
  n=f;
};

template<class T> inline void Array<T>::
Add(const T& i)
  { if(n>l) Error(ErrArrayOverflow);
    else arr[(n++)-f]=i;             };

template<class T> Iter<T,T,UWORD>& Array<T>::
Find(const T& c) const
  { for(UWORD q=0; q<(l-f+1); q++)
      if(arr[q]=c)
        return AIter<T>(*this,q+f);
    return AIter<T>(l-f+1);
  };

template<class T> void Array<T>::
Remove(const Iter<T,T,UWORD>& I)
  { I.ValidFor(*this); I.Valid();
    for(UWORD q=I.i; q<(l-f); q++)
      arr[q]=arr[q+1];
    n--;
  };
