// typedef enum {false, true} bool; //
#include <portab.h>
#include <error.h>
#include <alstring.h>

/* ------------------------------------------------------------------------ */
// FORWARD-Deklarationen
template<class I, class O> class Iter;
template<class T> class ArrayABC;
template<class T> class AIterABC;
template<class I, class O> class OIter;

/* ------------------------------------------------------------------------ */
// ContainerABC<I, O> & Iter<I, O> - abstract base class for EVERYTHING

template<class I, class O> class ContainerABC
{ public:
   virtual void        Add(const I&)   = 0;
   virtual O           GetRemove()     = 0;
   virtual const bool  IsEmpty()       = 0;
   virtual             ~ContainerABC()    { };

   ContainerABC& operator+=(const I& e)   { Add(e); return *this; };
   ContainerABC& operator<<(const I& e)   { Add(e); return *this; };
   O             operator() ()            { return GetRemove();   };
                 operator const bool()    { return !IsEmpty();    };

   ContainerABC() { };
  private:
    operator=(ContainerABC&); ContainerABC(ContainerABC&);
};

template<class I, class O, class BO, class E> class  Container
 /* I= Input Typ, O=Output type */         : public  ContainerABC<I,O>
 /* BO=base of O, E=Engine/Implementation*/: private E<BO>
{ public:
   void        Add(const I& a)   { E::Add(a);                };
   O           GetRemove()       { return (O)E::GetRemove(); };
   const bool  IsEmpty()         { return E::IsEmpty();      };

   Container& operator+=(const I& e)   { Add(e); return *this; };
   Container& operator<<(const I& e)   { Add(e); return *this; };
   O          operator() ()            { return GetRemove();   };
              operator const bool()    { return !IsEmpty();    };

   // Container() { };
  private:
    operator=(Container&); Container(Container&);
};


template<class I, class O> class Collection : public ContainerABC<I,O>
{ public:
   virtual O&          Get(Iter<I, O>& i)              { return i.Get(); };
   virtual const O&    Get(const Iter<I, O>& i) const  { return i.Get(); };
   virtual Iter<I, O>& Find(const I&) const            = 0;
   virtual void        Remove(const Iter<I, O>&)       = 0;
   virtual Iter<I, O>& NewIter() const                 = 0;

   O&       operator[](Iter<I, O>& i)                  { return Get(i);  };
   const O& operator[](const Iter<I, O>& i) const      { return Get(i);  };

  friend class Iter<I, O>;
};

template<class I, class O> class Iter
{   operator=(Iter&); Iter(Iter&);
  public:
    virtual bool     Valid() const                             = 0;
    virtual bool     ValidFor(const Collection<I, O>& c) const = 0;
    virtual const O& Get() const                               = 0;
    virtual       O& Get()                                     = 0;

    operator bool()                            { return Valid(); };
    virtual const O& operator() () const       { return Get();   };
    virtual O& operator() ()                   { return Get();   };

    Iter() { };
};

/* ------------------------------------------------------------------------ */
// Set<T>
// A Set is an unordered set (:-) of items; so you cannot iterate (++)
// through it!
template<class T> class Set : public Collection<T, T>
{ public:
}; // Q: Does this make sense??

/* ------------------------------------------------------------------------ */
// StackABC<T>
// A Stack is a stack. If you don't know what this is -> :-(
// (it's a Container that works after the LIFO principle)
template<class T> class StackABC : public ContainerABC<T, T>
{ public:
   void       Add(const T& a) = 0;
   T          GetRemove()     = 0;
   const bool IsEmpty()       = 0;
};

template<class T, class BT, class E> class  StackT2
 /* T=type of items, BT=base type */  : public  StackABC<T>
 /* E=Engine/Implementation       */  , private E
{ public:
   void        Add(const T& a)   { E::Add(a);                };
   T           GetRemove()       { return (T)E::GetRemove(); };
   const bool  IsEmpty()         { return E::IsEmpty();      };
};

//typedef template<class T, class E> /*class*/ Stack<T,void*,E> class StackP;
//template<class T, class E> typedef Stack<T,void*,E> class StackP;
#define StackP(T,E) StackT2<T,void*,E>
#define StackH(T,E) StackT2<T,HandleEngine,E>
#define StackT(T,E) StackT2<T,T,E>
/* ------------------------------------------------------------------------ */
// Queue<T>
// A Queue is a Container that works after the FIFO principle.
template<class T> class Queue : public ContainerABC<T, T>
{ public:
   void       Add(const T& a);
   T          GetRemove();
   const bool IsEmpty()        { return head==NIL; };

   Queue() : last(0), head(0)  { };
   ~Queue();

  private:
    Node<T>* last;
    Node<T>* head;
};

/* ------------------------------------------------------------------------ */
// OrdCollection<I, O> & OIter<I, O> - abstract base class
// An OrdCollection is an ordered thing: You can access the "third" item and
// you can iterate (by using UWORDs or OrdIters) through it.
// Attention: The order _always_ starts from one (1) on and goes up to a
// fixed or variable size which can be known be calling the Items() method.
// It is NOT allowed to derrive any class that allowes a range like [-5..3].
template<class I, class O> class OrdCollection : public Collection<I, O>
{  typedef OIter<I, O>& rOI;
  public:
   virtual O&          Get(const UWORD);
   virtual const O&    Get(const UWORD) const;
   virtual void        Remove(const UWORD r);
   virtual UWORD       Items() const             = 0;
           const O&    operator[](UWORD i) const { return Get(i);  };
           O& operator[](UWORD i)                { return Get(i);  };

   O           GetRemove() { return Get(Items()); Remove(Items()); };
   const bool  IsEmpty()   { return Items()==0;                    };

   O&       operator[](Iter<I, O>& i)                   { return i.Get();           }; // why must repeat?
   const O& operator[](const Iter<I, O>& i) const       { return i.Get();           }; // why must repeat?
   virtual  O&           Get(Iter<I, O>& i)             { return i.Get();           }; // why must repeat?
   virtual  const O&     Get(const Iter<I, O>& i) const { return i.Get();           }; // why must repeat?
   virtual  OIter<I, O>& NewOIter() const               { return TC(NewIter(),rOI); };
   virtual  OIter<I,O>&  Find_OIter(const I& f) const   { return TC(Find(f),rOI);   };
};

template<class I, class O> class OIter : public Iter<I, O>
{   operator=(OIter&); OIter(OIter&);
  public:
    virtual OIter& operator++ () = 0;
    virtual OIter& operator-- () = 0;
    virtual void   Set(UWORD p)  = 0;
    //virtual void operator++ (int) = 0;
    //void operator-- (int)         = 0;

    OIter() { };
};

/* ------------------------------------------------------------------------ */
// Array<T> & AIter<T>
// An Array behaves like a built-in array: It's size is fixed.

template<class T> class ArrayABC : public OrdCollection<T,T>
{ public:
   //ArrayABC(UWORD s);
   //virtual ~ArrayABC();

/*   virtual UWORD     Items() const              = 0;

   virtual void      Add(const T& i)            = 0;
   virtual T&        Get(const UWORD i)         = 0;
   virtual const T&  Get(const UWORD i) const   = 0;
   //virtual AIterABC<T>& Find_AIter(const T&) const = 0;
         OIter<T,T>& Find_OIter(const T& i) const = 0;
          Iter<T,T>& Find(const T& i) const { return (Iter<T,T>&)Find_OIter(i); }; // TC
   virtual void      Remove(const Iter<T,T>&)   = 0;
            OIter<T,T>& NewOIter() const         = 0;

  friend class AIterABC<T>;
*/};

/*
template<class T> class AIterABC : public OIter<T,T>
{  // operator=(AIterABC&); AIterABC(AIterABC&);
  public:
    //AIterABC(const ArrayABC<T>& c)              = 0;
    //AIterABC(const ArrayABC<T>& c, UWORD p)     = 0;
    //AIterABC& operator=(const AIterABC& s)      = 0;

/*  bool Valid() const                            = 0;
    bool ValidFor(const Collection<T, T>& c) const = 0;

    const T& Get() const                          = 0;
    T& Get()                                      = 0;

    //OIter<T,T>& operator++ ()                     = 0;
    //OIter<T,T>& operator-- ()                     = 0;
    void Set(UWORD p)                             = 0;

    friend class ArrayABC<T>;
*/}; */

template<class T> class Array1 : public ArrayABC<T>
{ public:
   void       Add(const T& i);
   T&         Get(const UWORD i)        { CheckBnd(i); return arr[i-1]; };
   const T&   Get(const UWORD i) const  { CheckBnd(i); return arr[i-1]; };
   Iter<T,T>& Find(const T&) const;
   void       Remove(const Iter<T,T>&);
   Iter<T,T>& NewIter() const           { return *new AIter1<T>(*this); };
   UWORD      Items() const             { return s;                     };

   Array1(UWORD s);
   Array1<T>& operator= (const Array1&);
   virtual ~Array1() { delete[] arr; };

  protected:
    UWORD s;    // size, valid Array indizes are in [1..s]
    UWORD n;    // next item position for a new item, valid range is [1..s]
    T*    arr;  // ptr to real C-array
    bool ValidBnd(UWORD j) const { return((j>0) && (j<n) ); };
    void CheckBnd(UWORD j) const { if(!ValidBnd(j)) Error(ErrArrayBounds); };
  friend class AIter1<T>;
};
#define Array Array1 // typedef doesn't work with templates

template<class T> class AIter1 : public OIter<T,T> // AIterABC<T>
{   void Checki() const { C.CheckBnd(i); };
  public:
    bool        Valid() const { return C.ValidBnd(i);                  };
    bool        ValidFor(const Collection<T, T>& c) const { if(&C!=&c) Error(ErrWrongIter); /*else*/ return true; };
    const T&    Get() const   { return C.Get(i);                       };
    T&          Get()         { return const_cast(Array<T>&,C).Get(i); };

    OIter<T,T>& operator++ () { Checki(); ++i; return *this;           };
    OIter<T,T>& operator-- () { Checki(); i--; return *this;           };
    void        Set(UWORD p)  { C.CheckBnd(p); i=p;                    };

    AIter1(const Array<T>& c)          : C(c) { i=1; };
    AIter1(const Array<T>& c, UWORD p) : C(c) { C.CheckBnd(p); i=p;  };
    AIter1(const AIter1& s)            : C(s.C), i(s.i)            { };
    AIter1& operator=(AIter1& s)       { ValidFor(s.C); i=s.i; s.i=0;
                                         delete &s; return *this;    };
  protected:
    const Array<T>& C;
    UWORD i;                  // currently "iterated" item; range [1..C.s]
  friend class Array1<T>;
};
#define AIter AIter1 // typedef doesn't work with templates

/* ------------------------------------------------------------------------ */
// List<T> & LIter<T>
// Lists are similar to Arrays, but they are "unlimited" - they never exceed.
// Lists have an additional method Insert() which does not override any item.

/* template<class T> class _ListABC : public OrdCollection<T, T>
{ public: virtual void Insert(const Iter<T,T>&) = 0; };

template<class T> class _List1 : public OrdCollection<T, T>, protected Array1<T>
{  typedef Array1<T> E // the Engine :-)
 public:
   void       Add(const T& i);
   T&         Get(const UWORD i)         { return E::Get(i); };
   const T&   Get(const UWORD i) const   { return E::Get(i); };
   Iter<T,T>& Find(const T& f) const     { return E::Find(f); };
   void       Remove(const Iter<T,T>& I) { E::Remove(I); };
   Iter<T,T>& NewIter() const            { return *new AIter1<T>(*this); };
   UWORD      Items() const              { return E::Items();            };
   void       Insert(const Iter<T,T>&)  ;

 protected:
	UWORD incr; // um wieviel Elemente wird size bei Bedarf erhht?
};*/

// PROBLEM: Is List a child of Array() ? It has the same interface; but the
// meaning changed a little... I don't know yet. I'll assume it is...
template<class T> class ListABC : public ArrayABC<T>
{public:
   virtual void Insert(const Iter<T,T>&, const T) = 0;
   virtual void Insert(const SWORD, const T);
};

template<class T> class List1 : virtual public ListABC<T>, virtual public Array1<T>
{ public:
   void Insert(const Iter<T,T>&, const T);
   void Insert(const SWORD, const T);

   void Add(const T& i);

   List1(UWORD iniSz=10, UWORD icr=5) : Array1<T>(iniSz), incr(icr) { };
   //List1(const List1& copyL);
   //List1& operator= (const List1& copyL);
   //~ListTpl();

 protected:
	UWORD incr; // um wieviel Elemente wird size bei Bedarf erhht?
};
#define List List1

/* ------------------------------------------------------------------------ */
// SortList<T> & sLIter<T>
// SortLists are similar to Lists, but all elements are automatically sorted
// Take care: You have no idea & control at which position an item is stored;
// but SortLists still are OrdCollections because you can iterate through them.
// REMARK: Though SortList has the same interface as List, it is not a List
// in the OOD way (in my opinion) because it behaves quite different.
template<class T> class SortListABC : public OrdCollection<T, T>
{public:
   virtual void Insert(const Iter<T,T>&) = 0;
};

/* ------------------------------------------------------------------------ */
// Dictionary<K,V> & DIter<V>
// A dictionary contains two lists: One for storing a _Key and another for a
// _Value. You have to use Add(K,V) to add a pair of a key and a value which
// you will be able to access later by using Find(K) with a key. Find() will
// return an Iter; use it to get back the Value with V Get(Iter).
template<class K, class V> class Dictionary : public OrdCollection<K, V>
{ public:
    void      Add(const K&, const &V);
    Iter<K,V> Find(const I&) const;
    V&        Get(const Iter<K, V>&);
    const V&  Get(const Iter<K, V>&) const;
    void      Remove(const Iter<L, V>&);

    void Add(const K&); // NOT defined! If used anyway, the linker detects it.
};

/* ------------------------------------------------------------------------ */
// OCollUWI<OrdCollection<I,O>&> - CollectionUserWorldInterface
// This FUNCTION (not method) converts ANY OrderedCollection into a string.
// Format: {item1, item2, ...}
template<class I, class O>
const String OCollUWI(OrdCollection<I,O>& C)
{ OIter<I, O>& J=C.NewOIter();
  String s("{"); String i;
  while(J)
    { s+=J(); ++J; if(J) s+=", "; };
  s+="}";
  delete &J;
  return s;
};
// Das obige soll eine allg. Prozedur zur Ausgabe von OrdCollection's sein.
// Geht aber nicht, weil der dmliche BC nicht schnallt, dass ein Array<char>
// ein OrdCollection<char,char> ist... (sollte er aber).
// Deshalb "von Hand" :
template<class T> const String OCollUWI(Array<T>& A)
{ typedef OrdCollection<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };

/* ------------------------------------------------------------------------ */
// ContUWI<ItemT,CollectionT> - CollectionUserWorldInterface
// Derrive any Collection class from ContUWI and it can be outputet to cout!
/*
template<class T> class ContUWI : virtual public OrdCollection<T,T>, public isASCIIable {
  String AsASCII() const
   { OIter<T,T,SWORD>& I=NewIter_OIter();
     String s("{");
     while(I)
       { s+=I(); I++; };
     s+="}";
     return s;             };
};
Das funktioniert aus div. Grnden (im Moment?) nicht wie gedacht...
*/

/* ------------------------------------------------------------------------ */
//                            IMPLEMENTATION
/* ------------------------------------------------------------------------ */
template<class I, class O> inline O& OrdCollection<I,O>::
Get(const UWORD r)
  { OIter<I,O>& I=NewOIter(); I.Set(r); O& o=Get(I); delete &I; return o; };

template<class I, class O> inline const O& OrdCollection<I,O>::
Get(const UWORD r) const
  { typedef const OIter<I,O>& crOI;
    OIter<I,O>& I=NewOIter(); I.Set(r);
    const O& o=Get(const_cast(crOI,I));
    delete &I; return o; };

template<class I, class O> inline void OrdCollection<I,O>::
Remove(const UWORD r)
  { OIter<I,O>& I=NewOIter(); I.Set(r); Remove(I); delete &I; };
/* ------------------------------------------------------------------------ */
template<class T> struct Node {
  T item;
  Node*  lnk;
  Node(T it, Node* n=NULL) : item(it), lnk(n) { };
};
/* ------------------------------------------------------------------------ */
// Stack<T>
// A Stack is a stack. If you don't know what this is -> :-(
// (it's a Container that works after the LIFO principle)

template<class T> class Stack_LL : public ContainerABC<T, T>
{ public:
   void       Add(const T& a);
   T          GetRemove();
   const bool IsEmpty()        { return last==NIL; };

   Stack_LL() : last(0), head(0)  { };
   ~Stack_LL();

  private:
    Node<T>* last;
    Node<T>* head;
};
//#define

template<class T> inline void Stack_LL<T>::
Add(const T& a)
{ Node<T>* N=new Node<T>(a);
  if(last)
    N->lnk=last;
  else
    head=N;
  last=N;
};

template<class T> inline T Stack_LL<T>::
GetRemove()
{ if(last) {
    T R=last->item;
    Node<T>* h=last->lnk;
    delete last;
    last=h;
    return R; }
  else
    Error(ErrContainerEmpty); 
};
template<class T> inline void Stack_LL<T>::
~Stack_LL()
{};

/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
template<class T> inline void Queue<T>::
Add(const T& a)
{ Node<T>* N=new Node<T>(a);
  if(last)
    last->lnk=N;
  else
    head=N;
  last=N;
};

template<class T> inline T Queue<T>::
GetRemove()
{ if(head) {
    T R=head->item;
    Node<T>* h=head->lnk;
    delete head;
    head=h;
    return R; }
  else
    Error(ErrContainerEmpty);
};

template<class T> inline Queue<T>::
~Queue()
{ while(head) {
    Node<T>* N=head->lnk;
    delete head;
    head=N;   };
};

/* ------------------------------------------------------------------------ */
template<class T> Array1<T>::
Array(UWORD size) {
  if(size<1)
    Error(ErrArrayBounds);

  arr=new T[s=size];
  n=1;
};

template<class T> inline Array1<T>& Array1<T>::
operator= (const Array1<T>& cpy)
  { if (cpy.s!=s) Error(ErrArrayBounds);
    for (n=0; n<cpy.n; n++)
      arr[n]=cpy.arr[n];
    return *this;
  };

template<class T> inline void Array1<T>::
Add(const T& i)
  { if(n>s) Error(ErrArrayOverflow);
    else arr[(n++)-1]=i;             };

template<class T> Iter<T,T>& Array1<T>::
Find(const T& c) const
  { for(UWORD q=0; q<(n-1); q++)
      if(arr[q]==c)
        return *new AIter1<T>(*this,q+1);
    AIter1<T>& NotValid=*new AIter1<T>(*this);
    NotValid.i=0;    // Ungltiger Iterator!
    return NotValid;
  };

template<class T> void Array1<T>::
Remove(const Iter<T,T>& I)
  { I.ValidFor(*this); I.Valid();                // Wenn I ValidFor dieses Array
    for(UWORD q=TC(I,AIter1<T>&).i-1; q<(n-2); q++) // ist, ist es sicher ein AIter1.
      arr[q]=arr[q+1];
    n--;
  };

/* ------------------------------------------------------------------------ */
template<class T> void ListABC<T>::
Insert(const SWORD r, const T it)
  { OIter<I,O>& I=NewOIter(); I.Set(r); Insert(I,it); delete &I; };

/* ------------------------------------------------------------------------ */
//template<class T> void List1<T>::
//Insert(const Iter<T,T>& i, const T it)



//template<class T> void List1<T>::
//Insert(const SWORD i, const T it)


//template<class T> void List1<T>::
//Add(const T& it)

