// handlecp.cpp - handle compact demo

#include <alhnd.h>
#include "handlecp.h"

void Demo0() {       // The old way: C-Pointer
   VarF* p1; VarF* p2; VarF* p3;

// 1. Creation
   p1=new(IntF(7));    p1->Print();    // shows: int 7
   p2=new(CharF('c')); p2->Print();    // shows: char c

// 2. Assignement
   // CRASH: p3->Print();
   p3=p2;              p3->Print();    // shows: char c
   // (*p3)='x';       This doesn't work, because the compiler doesn't know
   //                  that p3 actually points to a CharF.

   // Suppose the scope of p2 ends here. What to do?
   // delete p2;       // Should p2 be deleted?? (No, p3 still points to it.)

// 3. Changement, calling non-const member func
   p3->Incr();         p2->Print();    // shows: char d !
                                       // Obj 2 changed too when manipulated Obj 3

// 4. Assignement
   // delete p1;       // Should p1 be deleted?? (Yes!)
   p1=p2;              // address of former Obj 1 ist lost!!
   p1->Print();        // shows: char d!

// 5. Free'ing
   delete p3;          // Fine. Deleted Obj 2.
   delete p2;          // Wrong! Obj 2 was already deleted.
   delete p1;          // Wrong! Obj 2 was already deleted.
   // Obj 1 cannot be deleted because it's address is lost!
};

void Demo1() {      // A better way: SmartPointer
   SPtr<VarF> p1, p2, p3;

// 1. Creation
   p1=new(IntF(7));    p1->Print();    // shows: int 7
   p2=new(CharF('c')); p2->Print();    // shows: char c

// 2. Assignement
   // CRASH: p3->Print();
   p3=p2;              p3->Print();    // shows: char c
   // (*p3)='x';       This doesn't work, because the compiler doesn't know
   //                  that p3 actually points to a CharF.

   // Suppose the scope of p2 ends here. What to do?
   // Don't care about deleting...

// 3. Changement, calling non-const member func
   p3->Incr();         p2->Print();    // shows: char d !
                                       // Obj 2 changed too when manipulated Obj 3
// 4. Assignement
   p1=p2;              // p1 is autom. deleted!
   p1->Print();        // shows: char d!

// 5. Free'ing
   // Don't care about deleting...
   // This gives a compile-time error: delete p3; (because p3 is not really a pointer)
   // DON'T DO THAT: delete &(*p3);
};

void Demo2() { // Using Shells!

// 1. Creation
	MyVar V1=7;   V1.Print(); // shows: int 7
   MyVar V2='c'; V2.Print(); // shows: char c
	MyVar V3;

// 2. Assignement
	// RunTime-Error:   V3.Print();
	V3=V2;
	V3.Print(); // shows: char c
	V3='x';       V2.Print(); // shows: char c - Obj 2 did not change!

// 3. Changement, calling non-const member func
   V3=V2;        V3.Print(); // Obj 'x' is freed!
   V3.Incr();    V2.Print(); // shows: char c - Obj 2 did not change!
                 V3.Print(); // shows: 'd'

// 4. Assignement
   V1=V2;                    // Obj 1 is autom. deleted
   V1.Print();               // Shows: char c

// 6. Working with SPtr to Fruits
   SPtr<VarF> sp1=V1;

   sp1=V3;
   sp1->Print();
   cout<<endl;  // shows 'd'

	sp1->Incr();               // *sp1 is now 'e'
   V3.Print();                // Shows: char e - Obj 3 changed too when manipulated by pointer

// 7. const-Test

	const MyVar V4=V3;       // V4='e' but CONST

	sp1=V4;                  // COMPILER: Temporary used...
   sp1->Incr();
   sp1->Print();            // Shows: char 'f'
   V4.Print();              // Shows: char 'f' - CONST Obj 4 CHANGED!

// 5. Free'ing
// Don't have to free anything!

   // DON'T DO THAT: delete &(*sp1);
};

MyVar foo()
{
  //IntF R(9);       // WRONG!
  //MyVar R=IntF(9); // WRONG!
  MyVar R=*new IntF(9);

  return R;
};

int main()
{
  Demo2();


 // B. Using a static global Fruit.
   IntF I(5); I.MakeStatic(); // Without this line,
                              // I is attempted to be deleted at the end
   MyVar V5=I;
//   I.Incr();
   V5.Print();                // 5

   V5.Incr();
   I.Print(); cout << endl;   // 6; I changed! (oops)
/*
 // C. Return local static Fruit.
   V5=foo();
   V5.Print();               // Obejct refered to doesn't exist anymore!
*/

 // Working with Fruits directly: pointers to Fruits
//   MyVar V4(new CharF('K'));
/*
   VarF* pV=new CharF('p');
   MyVar V5=*pV;
   V5.Print();              // p

   V5.Incr();               // change on shell
   pV->Print();             // q, the pointed obj was Changed! (ooh)

   pV->Incr();              // change through fruit pointer
   V5.Print();              // r, the shell was changed! (ooh)

// DON'T: delete pV; The Shell does this for you!
*/

/*
 // Working with ShellEngine
   ShellEngine H1=V1;
   MyVar V4=(MyVar)H1;
   V4.Print();              // 7
*/
  return 0;
};
