// PROJECT : ALIB
// FILE    : DEBUG.CPP - implementation of debug facilities
// AUTHOR  : Michael Vorburger, alpha ware, CH-SG
// DATE oC : October 1994
// DATE oLM: October 1994
// LEVEL   : GLE

/* This module makes it possible to detect if memory on the heap was not
	free'd and who new'd it. This can be a great help to find memory problems. */

#include <debug.h>

#if defined(DEBUG)

#include <stdlib.h>
#include <iostream.h>
#include <msitext.h>
#include <mem_msi.h>

struct anAlloc {    // data keept of each allocation
   void*    p;          // where is it?
   unsigned s;          // which size has it? (eigt. size_t)
   char*    source;     // filename of source
   unsigned line;       // line in source
   // maybe add "time" and object name and ?
};

const unsigned sHCL=100;    // const size of the HCL array
unsigned       nHCL;        // next item in HCL
//unsigned       uHCL;        // used items in HCL
anAlloc        HCL[sHCL];   // array with all items. HCL=heap control list (?)
char*          HCnFile;     // argument passing new(T) macro to operator new()
unsigned       HCnLine;     // dito.
char*          LineUnknown="--unknown--";

/* REMARK: all memory allocations are stored in the HCL array. This array is
   used as a "ring" - it's cyclic. Rings could be implemented with modulo %
   ...but I did it differnt (mainly because at first I forgot about the
   modulo method and then I was too lazy to change the code again.          */

void InitHCL()
{ /*uHCL=0*/
   for(nHCL=0; nHCL<sHCL; nHCL++)
     HCL[nHCL].p=NULL;
   nHCL=0;
};

#undef new
void* operator new(size_t s)  {
  void* t;
  if(s) {
    t=malloc(s);
    if (t) {
      unsigned stop=nHCL-1;
		if(nHCL==0) stop=sHCL-1;
      while(HCL[nHCL].p && nHCL!=stop)
        if(nHCL<(sHCL-1)) ++nHCL; else nHCL=0;
      if(nHCL==stop && HCL[nHCL].p) NativePrint("HCL is FULL. NO MORE HEAP CHECKING... increase sHCL");
      else {
        HCL[nHCL].p=t;
        HCL[nHCL].s=s;
        if(HCnFile==0) HCnFile=LineUnknown;
        HCL[nHCL].source=HCnFile; HCnFile=NULL;
        HCL[nHCL].line=HCnLine;   HCnLine=0;
        if(++nHCL==sHCL) nHCL=0;
      };
    }
    else
		{ NativePrint(CR); NativePrint("No more memory! [FATAL]");
        NativePrint(CR); abort();                                };
  } else
      { NativePrint(CR); NativePrint("Attempt to allocate (new) zero (0) bytes.");
        NativePrint(CR); /* don't abort. */                                 };
  return t;
};
#define new(T) alpha_new(T)

void operator delete(void* p) {
  if ( heapchecknode(p) != _USEDENTRY )
	 { if (TxtCursX()>1) NativePrint(CR);
		NativePrint("Attempt do delete a block that is "
						"already free or was never new'd." );
		NativePrint(CR);
	 };
  unsigned stop=nHCL+1;
  if(nHCL==sHCL-1) stop=0;
  while(HCL[nHCL].p!=p && nHCL!=stop)
	 if(nHCL!=0) --nHCL; else nHCL=sHCL-1;
  if(HCL[nHCL].p==p) // means: found
	 { free(p); HCL[nHCL].p=NULL; }
};

inline void ForgotToFree(unsigned i)
  { // printf("FORGOT TO free() %u BYTES at %p // new()'d IN FILE <%s> #%u\n",HCL[i].s,HCL[i].p,HCL[i].source,HCL[i].line); };
	 cerr << "FORGOT TO free() "<<HCL[i].s<<" BYTES at "<<HCL[i].p<<" // new()'d IN FILE <"
			<<HCL[i].source<<"> #"<<HCL[i].line<<endl; };

void CloseHCL()
{ unsigned stop=nHCL+1;
  if(nHCL==sHCL-1) stop=0;
  while(nHCL!=stop)
    { if(HCL[nHCL].p) { ForgotToFree(nHCL); delete HCL[nHCL].p; };
      if(nHCL!=0) --nHCL; else nHCL=sHCL-1; };
  if(HCL[nHCL].p) // check last item too.
    { ForgotToFree(nHCL); delete HCL[nHCL].p; };
};

class DebugMain
{ ULONG HeapMem;
 public:
  DebugMain()
  { /*clrscr();*/ // HeapMem=coreleft();
    InitHCL();
    cerr << ">>>Program started>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << endl;
  };
  ~DebugMain()
  { // signed long diff=HeapMem-coreleft();
    // ?? [UNZUVERLSSIG] 48 bytes diff verschwinden offenbar wg. Klasse
    // ?? 32 sind es, wenn mind. eine Klasse mit virtuellen Methoden
    //if( (diff!=48) && (diff!=32) )
    //  NativePrint("Memory lost!"); NativePrint(CR);
    // cerr << "Memory lost! Diff=" << diff << " bytes" << endl;
    if(TxtCursX()>1) cerr << endl;
    cerr << "===============================================================Program ended===" << endl;
    CloseHCL(); };
} debugMain;


void ForgotToLinkDebugCPP() { }; // see DEBUG.H


/* --------------------------------------------------------------------*/


debugable::debugable()
{
        cerr << "CONSTRUCT: new object of class >";
// siehe unten zum folgenden...
/*      if ( TypeOf() < IdAppClassesFirst )
          { cerr << classNamesAlpha[TypeOf()]; }
        else
          cerr << classNamesApp[TypeOf()];
*/      cerr << "< with DebugId #" << (number=nextnumber++)
             << " (at adr " << this << ")" << endl;
};

debugable::~debugable()
{       cerr << "DESTRUCT: object #" << number
             << " (at adr " << this << ")" << endl;
};

UWORD debugable::nextnumber = 1;
// MEMSIZE debugable::allocatedObjMem = 0; // Wieviel Heap-Speicher wird durch
// Objekte belegt? Muss wieder Null sein am Ende.


#endif //  defined(DEBUG)
