// PROJECT : TERM
// FILE    : headerfile and some tiny inline implementations
// AUTHOR  : Michael Vorburger, alpha ware
// DATE oC : July 1994
// DATE oLM: August 1994

#ifndef __TERM_H
#define __TERM_H

#include <alobject.h>
#include <error.h>
#include <alstring.h>
#include <list.h>
#include "ids_term.h"   // must be last #include

//class TermABC;
//class MultiplyClass;
//class PlusClass;
//class PowerClass;

enum CalcType { std, sym, num };

#define const const

class hTerm;
class Numerical;
class Double;   extern Double StdNumerical;
class T_NoTermF; extern T_NoTermF NoTerm;

class TermABC : public Fruit, public isASCIIable //, public debugable
{ public:
    // FRUITMACRO(TermABC) not neccessary for ABC

    virtual hTerm Plus(const TermABC& t, CalcType ct=std)     const = 0;
    virtual hTerm Multiply(const TermABC& t, CalcType ct=std) const = 0;
    virtual hTerm Power(const TermABC& t, CalcType how=std)   const = 0;

// These methods are the same as above but apply to YourSelf; they change the
// object itself instead returning a new new. This is good for performance.
//  virtual hTerm PlusYS(const TermABC& t, CalcType ct=std)     = 0;
//  virtual hTerm MultiplyYS(const TermABC& t, CalcType ct=std) = 0;
//  virtual hTerm PowerYS(const TermABC& t, CalcType how=std)   = 0;

    //virtual hTerm ReCalc(TermABC& t, CalcType how)  const;
    virtual hTerm Numeric(Numerical& N) const;

    virtual bool operator== (const TermABC& c) const { return true;  };
// not needed: virtual bool operator!= (const TermABC& c) const { return !operator==(c); };
    virtual bool operator>  (const TermABC& c) const { return false; };
    virtual bool operator<  (const TermABC& c) const { return !operator>(c);  };
    virtual bool operator>= (const TermABC& c) const { return operator>(c) || operator==(c); };
    virtual bool operator<= (const TermABC& c) const { return operator<(c) || operator==(c); };

// inherited:
    virtual operator bool()    const { return true; };
    virtual String AsASCII()   const = 0;
    virtual BOOL IsA(UWORD id) const
        { return( id==IdTermABC
//                     || isASCIIable::IsA(id)
//                     || isSortable::IsA(id)
                );
        };
    virtual UWORD TypeOf()     const { return IdTermABC; };
};

class Integer;
class Variable;

// hTerm ist das Handle fr Term und wird intern verwendet. Siehe unten.
class hTerm : public Shell<TermABC>
{ public:
    SHELL_INTERFACE(hTerm, TermABC)
    DEC_FRUITINIT_INTERFACE_FAM(hTerm,Integer,TermABC, SWORD) // S F T
    DEC_FRUITINIT_INTERFACE_FAM(hTerm,Real,TermABC,double)
    DEC_FRUITINIT_INTERFACE_FAM(hTerm,Variable,TermABC,String)
    operator const TermABC&() const { return Frt(); };
    //operator bool() const { return Shell<TermABC>::operator bool(); };

/* STOP: These operators are only supported in Term. This internal Class works
         with the corresponding Plus/Multiply/... functions!
    hTerm operator+ (const hTerm& t) const { return Frt().Plus(t.Frt() ); };
    hTerm operator- (const hTerm& t) const; // { return( operator+ ( t*Integer_1 ) ); };
    hTerm operator* (const hTerm& t) const { return Frt().Multiply(t.Frt(), std); };
    hTerm operator^ (const hTerm& t) const { return Frt().Power(t.Frt(), std); };
    hTerm operator/ (const hTerm& t) const; // { return( operator* ( t^Integer_1 ) ); };

// later we will introduce TermABC::PlusYS etc. and the change this
    hTerm operator+= (const hTerm& t)  { EqualFrtPtr(&( Frt().Plus(t.Frt())    .Frt() )); return *this; };
    hTerm operator-= (const hTerm& t)  { EqualFrtPtr(&( operator-(t)           .Frt() )); return *this; };
    hTerm operator*= (const hTerm& t)  { EqualFrtPtr(&( Frt().Multiply(t.Frt()).Frt() )); return *this; };
    hTerm operator^= (const hTerm& t)  { EqualFrtPtr(&( Frt().Power(t.Frt())   .Frt() )); return *this; };
    hTerm operator/= (const hTerm& t)  { EqualFrtPtr(&( operator/(t)           .Frt() )); return *this; };
*/
    hTerm Plus(const TermABC& t, CalcType ct=std)     const { return Frt().Plus(t,ct); };
    hTerm Multiply(const TermABC& t, CalcType ct=std) const { return Frt().Multiply(t,ct); };
    hTerm Power(const TermABC& t, CalcType ct=std)    const { return Frt().Power(t,ct); };

// operators== and != are supported by Handle
//    bool operator== (const hTerm& c) const { return Frt().operator==(c.Frt()); };
//    bool operator!= (const hTerm& c) const { return Frt().operator!=(c.Frt()); };
    bool operator>  (const hTerm& c) const { return Frt().operator> (c.Frt()); };
    bool operator<  (const hTerm& c) const { return Frt().operator< (c.Frt()); };
    bool operator>= (const hTerm& c) const { return Frt().operator>=(c.Frt()); };
    bool operator<= (const hTerm& c) const { return Frt().operator<=(c.Frt()); };

    hTerm Numeric(Numerical& N) const  { return Frt().Numeric(N); };

    String AsASCII() const { return Frt().AsASCII(); };

    friend class Term;     // damit operatoren auf t.Frt() zugreifen knnen
};

// Term ist von hTerm abgeleitet und diejenige Klasse, mit die der Benutzer
// verwendet. Notwendig weil a) AsASCII das usserste Klammerpaar entfernt
// in _diese_ Klasse spter ein catch(...) reinkommt!
class Term : public hTerm, public isASCIIable
{ public:
    DERR_SHELL_INTERFACE(Term,hTerm,TermABC)
    DEC_DERRSH_FRTINI_INTF_FAM(Term,hTerm,Integer,TermABC,SWORD)
    DEC_DERRSH_FRTINI_INTF_FAM(Term,hTerm,Real,TermABC,double)
    DEC_DERRSH_FRTINI_INTF_FAM(Term,hTerm,Variable,TermABC,String)

// These operators HAVE TO return Term instead hTerm, so that cout << a^2 is possible
    Term operator+ (const hTerm& t) const { return Frt().Plus(t.Frt() ); };
    Term operator- (const hTerm& t) const; // { return( operator+ ( t*Integer_1 ) ); };
    Term operator* (const hTerm& t) const { return Frt().Multiply(t.Frt(), std); };
    Term operator^ (const hTerm& t) const { return Frt().Power(t.Frt(), std); };
    Term operator/ (const hTerm& t) const; // { return( operator* ( t^Integer_1 ) ); };

// later we will introduce TermABC::PlusYS etc. and the change this
    hTerm operator+= (const hTerm& t)  { EqualFrtPtr(&( Frt().Plus(t.Frt())    .Frt() )); return *this; };
    hTerm operator-= (const hTerm& t)  { EqualFrtPtr(&( operator-(t)           .Frt() )); return *this; };
    hTerm operator*= (const hTerm& t)  { EqualFrtPtr(&( Frt().Multiply(t.Frt()).Frt() )); return *this; };
    hTerm operator^= (const hTerm& t)  { EqualFrtPtr(&( Frt().Power(t.Frt())   .Frt() )); return *this; };
    hTerm operator/= (const hTerm& t)  { EqualFrtPtr(&( operator/(t)           .Frt() )); return *this; };

    //hTerm Numeric(Numerical& N) const;

    String AsASCII() const
      { String s=Frt().AsASCII();
        if (s[0]=='(') { s.Mid(1,s.Length()-2); };
        return s;                                 };
};

class T_NoTermF : public TermABC
{ public:
    hTerm Plus(const TermABC& t, CalcType ct=std) const     { return NoTerm; };
    hTerm Multiply(const TermABC& t, CalcType ct=std) const { return NoTerm; };
    hTerm Power(const TermABC& t, CalcType ct=std) const    { return NoTerm; };
/*
    virtual bool IsZero() const ;
    virtual bool IsPositive() const;
    virtual bool IsOne() const;
*/
    //hTerm ReCalc(const TermABC& t, CalcType how)  const

    operator bool()  const { return false; };
    String AsASCII() const { return "--no hTerm built/found/???--"; };

    FRUITMACRO(T_NoTermF)
    BOOL  IsA(UWORD id) const { return( id==IdINVTerm || TermABC::IsA(id)); };
    UWORD TypeOf()      const { return IdINVTerm; };
};

class T_ErrTermF;
extern T_ErrTermF ErrTerm;
class T_ErrTermF : public T_NoTermF
{ public:
    hTerm Plus(const TermABC& t, CalcType ct=std) const     { return ErrTerm; };
    hTerm Multiply(const TermABC& t, CalcType ct=std) const { return ErrTerm; };
    hTerm Power(const TermABC& t, CalcType ct=std) const    { return ErrTerm; };

    operator bool()  const { return true;        }; // true is fine. NOT false.
    String AsASCII() const { return "--ERROR--"; };
};

/* -------------------------------------------------------------------*/

class Operator : public TermABC
{public:
  hTerm Plus(const TermABC& e, CalcType how)     const  { return TermABC::Plus(e, how); };
  hTerm Multiply(const TermABC& e, CalcType how) const  { return TermABC::Multiply(e, how); };
  hTerm Power(const TermABC& e, CalcType how)    const  { return TermABC::Power(e, how); };

  BOOL IsA(UWORD id) const { return( id==IdOperator || TermABC::IsA(id) ); };
  UWORD TypeOf()     const { return IdOperator; };
};

class Operand : public TermABC
{public:
  hTerm Plus(const TermABC& e, CalcType how)     const { return TermABC::Plus(e, how); };
  hTerm Multiply(const TermABC& e, CalcType how) const { return TermABC::Multiply(e, how); };
  hTerm Power(const TermABC& e, CalcType how)    const { return TermABC::Power(e, how); };

  BOOL IsA(UWORD id) const    { return( id==IdOperand || TermABC::IsA(id) ); };
  UWORD TypeOf()     const    { return IdOperand; };
};

class Real;

class Numerical : public Operand
{public:
//  hTerm NPow_(Numerical& e);
//  virtual Numerical& NPow(Numerical& e) = 0;
//  hTerm NMultiply_();
  hTerm Plus(const TermABC& e, CalcType how)         const;
  hTerm Multiply(const TermABC& e, CalcType how=std) const;
  hTerm Power(const TermABC& e, CalcType how=std)    const;

  virtual bool IsZero()     const = 0;
  virtual bool IsPositive() const = 0;
  bool    IsNegative()      const { return ( (!IsZero()) && (!IsPositive()) ); };
  virtual bool IsOne()      const = 0;
//
  virtual operator double() const = 0;
  virtual operator SWORD()  const = 0;

  virtual void Set(const Real&)    = 0;
  virtual void Set(const Integer&) = 0;

  Numerical& operator=(const TermABC& N)
   { if(N.IsA(IdReal) )
       Set(TC(N,const Real&));
     else if(N.IsA(IdInteger))
       Set(TC(N,const Integer&));
     return *this;
   };
//
  BOOL IsA(UWORD id) const  { return( id==IdNumerical || TermABC::IsA(id) ); };
  UWORD TypeOf()     const  { return IdNumerical; };
};

class Natural : public Numerical
{public:
  BOOL IsA(UWORD id) const  { return( id==IdNatural || Numerical::IsA(id) ); };
  UWORD TypeOf()     const  { return IdNatural; };
};

const char ApproxChar='';  // sollte konstante innerhalb der Klasse sein.
class Approximate : public Numerical
{public:
  BOOL IsA(UWORD id) const { return( id==IdApproximate || Numerical::IsA(id) ); };
  UWORD TypeOf()     const { return IdApproximate; };
};

class Special : public Numerical
  {};

class Imaginary : public Special
  {};

class Infinite : public Special
  {};

/* -------------------------------------------------------------------*/

class Integer : public Natural
{ private:
    SWORD i;

  public:
    Integer(SWORD v)  { i=v; };

    hTerm Plus(const TermABC& t, CalcType ct = std)     const;
    hTerm Multiply(const TermABC& t, CalcType ct = std) const;
    hTerm Power(const TermABC& t, CalcType ct = std)    const;

//    virtual Numerical& NPow(Numerical& e);

    bool IsZero()      const  { return( i==0 ); };
    bool IsPositive()  const  { return( i>0 );  };
    bool IsOne()       const  { return( i==1 ); };

    operator double()  const  { return i;       };
    operator SWORD()   const  { return i;       };
    void Set(const Real& R)   ;
    void Set(const Integer& I){ i=I;            };

    FRUITMACRO(Integer)
    String AsASCII()   const  { return(i);      };
    BOOL IsA(UWORD id) const  { return( id==IdInteger || Numerical::IsA(id) ); };
    UWORD TypeOf()     const  { return IdInteger; };
};

class Real : public Approximate
{ private:
    double d;

  public:
    Real(double i) : d(i) {};

    bool IsZero()      const  { return( d==0 ); };
    bool IsPositive()  const  { return( d>0 );  };
    bool IsOne()       const  { return( d==1 ); };

    hTerm Plus(const TermABC& t, CalcType ct = std)     const;
    hTerm Multiply(const TermABC& t, CalcType ct = std) const;
    hTerm Power(const TermABC& t, CalcType ct = std)    const;

    operator double()  const  { return d;       };
    operator SWORD()   const  { return d;       };
    void Set(const Real& R)   { d=R;            };
    void Set(const Integer& I);

    FRUITMACRO(Real)
    String AsASCII()   const  { String s(ApproxChar); s.SetFloatFormat("%g");
                                s<<d; return(s);                              };
    BOOL IsA(UWORD id) const  { return( id==IdReal || Approximate::IsA(id) ); };
    UWORD TypeOf()     const  { return IdReal; };
};

class Variable : public Operand     // NOT Numerical!
{
  private:
    String id; // identifier
//  hTerm   ct; // contents
  public:
    Variable(const String& name /* ,hTerm init = NIL */) : id(name) // , ct(init)
      { };

    virtual hTerm Plus(const TermABC& t, CalcType ct = std)     const;
    virtual hTerm Multiply(const TermABC& t, CalcType ct = std) const;
    virtual hTerm Power(const TermABC& t, CalcType ct = std)    const;

    virtual String AsASCII() const
      { return id; };

    FRUITMACRO(Variable)
    virtual UWORD TypeOf() const
      { return IdVariable; };
    BOOL IsA(UWORD id) const
      { return( id==IdVariable || Operand::IsA(id) ); };
};

class PlusClass : public Operator
{  private:
     hTerm s1;  // 1. Summand
     hTerm s2;  // 2. Summand

   public:
    PlusClass(const TermABC& a, const TermABC& b)
      : s1(a), s2(b)
      { };

    virtual hTerm Plus(const TermABC& t, CalcType ct = std)     const;
    virtual hTerm Multiply(const TermABC& t, CalcType ct = std) const;
    virtual hTerm Power(const TermABC& t, CalcType ct = std)    const;

    virtual String AsASCII() const
        { return "("+s1.AsASCII()+"+"+s2.AsASCII()+")"; };
        // Tip: "+" statt '+' umgeht ambiguity-Problem der String-Klasse!

    FRUITMACRO(PlusClass)
    virtual BOOL IsA(UWORD id) const
        { return( id==IdPlus || Operator::IsA(id) ); };
    virtual UWORD TypeOf() const
        { return IdPlus; };
};

class MultiplyClass : public Operator
{ private:
    SortList_A<hTerm> L;

  public:
    MultiplyClass(const TermABC& a, const TermABC& b)
      { L+=(a); L+=(b); };

    hTerm Plus(const TermABC& t, CalcType ct = std)     const;
    hTerm Multiply(const TermABC& t, CalcType ct = std) const;
    hTerm Power(const TermABC& t, CalcType ct = std)    const;

    String AsASCII() const;

    FRUITMACRO(MultiplyClass)
    virtual BOOL IsA(UWORD id) const
        { return( id==IdMultiply || Operator::IsA(id) ); };
    virtual UWORD TypeOf() const
        { return IdMultiply; };
};

class PowerClass : public Operator
{ private:
    hTerm b;  // Basis
    hTerm e;  // Exponent

  public:
    //PowerClass(const TermABC& base, const TermABC& exponent)
    PowerClass(const hTerm base, const hTerm exponent)
      : b(base), e(exponent) { };

    hTerm Plus(const TermABC& t, CalcType ct = std)     const;
    hTerm Multiply(const TermABC& t, CalcType ct = std) const;
    hTerm Power(const TermABC& t, CalcType ct = std)    const;

    //hTerm ReCalc(Numerical& t, CalcType how) const
    hTerm Numeric(Numerical& N) const
      { N=b;
        return N.Power(e,num); };

    String AsASCII()   const { return "("+b.AsASCII()+"^"+e.AsASCII()+")"; };

    FRUITMACRO(PowerClass)
    BOOL IsA(UWORD id) const { return( id==IdPower || Operator::IsA(id) ); };
    UWORD TypeOf()     const { return IdPower; };
};


/* -------------------------------------------------------------------*/
//  some INLINE Implementations
// =============================

// -> most of this is in Term.CPP
// but inlines can only be defined in .H files (?)

#endif // __TERM_H
